/*
** ocd - ojnk cd player
** Copyright (C) 1999 Odin <odin@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
*/

#include <ocd.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <sys/stat.h>

static u_long select_inexact(const struct inexact *ie, u_char len, u_char
*genre, size_t n, u_char);
static FILE *cddb_opendb(const u_char *filename, const u_char *mode);

extern int max_x, max_y;
WINDOW *cwin;
static u_char c_sel;

#define IE_FIRST (6)
#define IE_LAST  (max_y - 2)
#define IE_LINES ((IE_LAST - IE_FIRST) + 1)

#define CL_FIRST (6)
#define CL_LAST  (max_y - 4)
#define CL_LINES ((CL_LAST - CL_FIRST) + 1)

void cddb_control(struct cdrom *cdrom) {
	u_char filename[PATH_MAX], *orig = NULL, gbuf[24];
	struct stat st;
	FILE *fp;

	c_sel = 1;

	if (cwin == NULL) {
		cwin = newwin(0, 0, 0, 0);
		keypad(cwin, TRUE);
	}

	wclear(cwin);
	snprintf(filename, sizeof(filename), "%s/%08lx",
		cdrom->cddb.dir, cdrom->cddb.discid);

	for (;;) {
		cddb_window(&cdrom->cddb);
		switch (wgetch(cwin)) {
			/* Move down the list of tracks */
			case KEY_DOWN:
				if (INVALID_TRACK(c_sel + 1, cdrom->cddb.tracks)) {
					break;
				}
				++c_sel;
				break;
			/* Move up the list of tracks */
			case KEY_UP:
				if (INVALID_TRACK(c_sel - 1, cdrom->cddb.tracks)) {
					break;
				}
				--c_sel;
				break;
			/* Move to bottom of track list */
			case KEY_END:
				c_sel = cdrom->cddb.tracks;
				break;
			/* Move to start of track list */
			case KEY_HOME:
				c_sel = 1;
				break;
			/* Page Up, up 5 tracks */
			case KEY_PPAGE:
				if (INVALID_TRACK(c_sel - 5, cdrom->cddb.tracks))
					c_sel = 1;
				else
					c_sel -= 5;
				break;
			/* Page down, down 5 tracks */
			case KEY_NPAGE:
				if (INVALID_TRACK(c_sel + 5, cdrom->cddb.tracks))
					c_sel = cdrom->cddb.tracks;
				else
					c_sel += 5;
				break;
			case '\f':
				wclear(cwin);
				break;
			case 'Q':
			case 'q':
//			c_quit:
				wclear(cwin);
				return;
			case 'c':
			case 'C':
//			c_clear:
				fp = cddb_opendb(filename, "w+");
				if (fp == NULL)
					return;
				cddb_unknown(fp, &cdrom->disc, cdrom->cddb.discid);
				fseek(fp, 0, SEEK_SET);
				cddb_read(fp, &cdrom->cddb);
				break;
			case 'D':
			case 'd':
//			c_download:
				mvwaddstr(cwin, CDDB_STWIN_Y, CDDB_STWIN_X,
					"                                              ");
				if (lstat(filename, &st) != -1) {
					orig = strdup(filename);
					if (orig == NULL) {
						endwin();
						fprintf(stderr, "Fatal: Out of memory.\n");
						exit(-1);
					}
					xstrncat(filename, ".bak", sizeof(filename));
				}

				if (rename(orig, filename) == -1)
					mvwprintw(cwin, CDDB_STWIN_Y, CDDB_STWIN_X,
						"Error: Cannot rename %s to %s.", orig, filename);

				fp = cddb_opendb((orig ? orig : filename), "w+");
				if (fp == NULL)
					return;
				if (cddb_get(&cdrom->disc, &cdrom->cddb, fp,
						cdrom->cddb.server, cdrom->cddb.port) == -1) {
					if (orig != NULL) {
						fclose(fp);
						rename(filename, orig);
						xstrncpy(filename, orig, sizeof(filename));
						xfree(orig);
						break;
					} else
						cddb_unknown(fp, &cdrom->disc, cdrom->cddb.discid);
				}

				if (orig != NULL) {
					unlink(filename);
					xstrncpy(filename, orig, sizeof(filename));
					xfree(orig);
				}

				fseek(fp, 0, SEEK_SET);
				cddb_read(fp, &cdrom->cddb);
				break;
			case 'R':
			case 'r':
//			c_reread:
				fp = cddb_opendb(filename, "r");
				if (fp != NULL)
					cddb_read(fp, &cdrom->cddb);
				break;
			case 'S':
			case 's':
				wmove(cwin, CDDB_STWIN_Y, CDDB_STWIN_X);
				wclrtoeol(cwin);
				waddstr(cwin, "Enter a genre: ");
				wrefresh(cwin);
				user_input(cwin, CDDB_STWIN_Y, CDDB_STWIN_X + 16, 0,
						gbuf, sizeof(gbuf));
				if (cddb_valid_genre(gbuf)) {
					cddb_submit(&cdrom->cddb, gbuf,
						cdrom->cddb.sserver, cdrom->cddb.sport);
				} else {
					wmove(cwin, CDDB_STWIN_Y, CDDB_STWIN_X);
					wprintw(cwin, "Error: Invalid genre: \"%s\"", gbuf);
				}
				break;
			case '!':
				mvwaddstr(cwin, 5, 5, cdrom->cddb.ext);
				break;
		}
	}
}

void cddb_window(cddb_t *cddb) {
	u_char *p, *tbuf;
	u_char *ttitle = xcalloc(sizeof(cddb->title), sizeof(u_char));
	u_char y = 3, track, i;

	if (cddb != NULL) {
		xstrncpy(ttitle, cddb->title, sizeof(cddb->title));
		wattrset(cwin, 0);
		p = strchr(ttitle, '/');

		if (p == NULL) {
			free(ttitle);
			mvwprintw(cwin, y++, 2, "Artist / Title: %s", cddb->title);
		} else {
			*p++ = '\0';
			while (isspace(*p))
				*p++ = '\0';
			mvwprintw(cwin, 1, 11, "%08lx", cddb->discid);
			mvwaddstr(cwin, 2, 11, ttitle);
			mvwaddstr(cwin, 3, 11, p);
			free(ttitle);
		}
	}

	wmove(cwin, 0, 0);
	wattrset(cwin, A_BOLD | COLOR_PAIR(8));
	whline(cwin, ACS_HLINE, max_x);

	wmove(cwin, 0, 3);
//	wattroff(cwin, A_BOLD | COLOR_PAIR(8));
	waddstr(cwin, "[");
	wattrset(cwin, A_BOLD);
	waddstr(cwin, "CDDB Control");
	wattrset(cwin, A_BOLD | COLOR_PAIR(8));
	waddstr(cwin, "]");

	boldword(cwin, 1, 2, "d", "iscid", " - ", 6);
	boldword(cwin, 2, 2, "a", "rtist", " - ", 6);
	boldword(cwin, 3, 3, "t", "itle", " - ", 6);
	boldword(cwin, max_y - 3, 3, "s", "tatus", "", 6);
	boldword(cwin, 5, 3, "t", "rack list", "", 6);

	wattrset(cwin, A_BOLD | COLOR_PAIR(8));
	wmove(cwin, max_y - 3, 0);
	wvline(cwin, ACS_VLINE, 3);
	whline(cwin, ACS_HLINE, 3);
	wmove(cwin, max_y - 1, 0);
	whline(cwin, ACS_HLINE, 54);

	wmove(cwin, 5, 0);
	whline(cwin, ACS_HLINE, 3);
	wvline(cwin, ACS_VLINE, max_y - 2);
	whline(cwin, ACS_ULCORNER, 1);
	mvwhline(cwin, max_y - 3, 0, ACS_ULCORNER, 1);
	mvwhline(cwin, max_y - 1, 0, ACS_LLCORNER, 1);

	if (c_sel > CL_LINES)
		track = c_sel - CL_LINES;
	else
		track = 0;

	wattrset(cwin, 0);
	tbuf = xcalloc(cddb->llen - 7, sizeof(u_char));
	y = CL_FIRST;

	for (i = 0 ; i < CL_LINES && track < cddb->tracks ; track++, i++) {
		if (track + 1 == c_sel)
			wattrset(cwin, ATTR_CURSEL);
		else if (track == c_sel)
			wattrset(cwin, 0);
		snprintf(tbuf, cddb->llen - 7, "%+2u - %s", track + 1, cddb->track[track].title);
		mvwprintw(cwin, y++, 2, "%-*s", cddb->llen - 8, tbuf);
	}

	wattrset(cwin, 0);
	wrefresh(cwin);
}

static FILE *cddb_opendb(const u_char *filename, const u_char *mode) {
	FILE *fp = fopen(filename, mode);

	if (fp == NULL) {
		mvwaddstr(cwin, CDDB_STWIN_Y, CDDB_STWIN_X,
			"Error: Cannot open CDDB database file.");
		beep();
		sleep(1);
		return (NULL);
	}

	return (fp);
}

/*
int cddb_edit(cddb_t *cddb) {
}
*/

u_long cddb_inexact(int sock, u_char *genre, size_t len) {
	u_char buf[128], *p, *q, i;
	struct inexact inexact[56];
	u_char longl = 0, temp;

	for (i = 0 ; i < (sizeof(inexact) / sizeof(struct inexact)) && sock_readline(sock, buf, sizeof(buf)) ; i++) {
		if (buf[0] == '.')
			break;
		p = strchr(buf, ' ');
		if (p == NULL)
			continue;
		*p++ = '\0';
		q = strchr(p, ' ');
		if (q == NULL)
			continue;
		*q++ = '\0';

		inexact[i].discid = strtoul(p, NULL, 16);
		p = strchr(q, '\r');
		if (p != NULL)
			*p = '\0';
		xstrncpy(inexact[i].genre, buf, sizeof(inexact[i].genre));
		xstrncpy(inexact[i].title, q, sizeof(inexact[i].title));
		temp = strlen(q);
		if (temp > longl)
			longl = temp;
	}

	/* Drop excess matches */
	while (buf[0] != '.')
		sock_readline(sock, buf, sizeof(buf));

	return (select_inexact(inexact, i, genre, len, longl));
}

static u_long select_inexact(const struct inexact *ie, u_char len, u_char *genre, size_t n, u_char longl) {
	u_char y, isel = 0, i, j;
	u_char tmp[120];
	WINDOW *iwin = newwin(0, 0, 0, 0);

	iwin = newwin(0, 0, 0, 0);
	keypad(iwin, TRUE);
	wclear(iwin);

	wattrset(iwin, A_BOLD);
	mvwaddstr(iwin, 0, 32, "Inexact/Multiple Matches");
	wattrset(iwin, 0);
	mvwaddstr(iwin, 2, 2,
		"Select the appropriate disc or press 'q' if not listed");
	wattron(iwin, A_UNDERLINE);
	mvwaddstr(iwin, 5, 5, "DISCID\tARTIST / TITLE");

	for (;;) {
		wmove(iwin, 6, 0);
		wattrset(iwin, 0);

		if (isel > IE_LINES)
			i = isel - IE_LINES;
		else
			i = 0;

		y = IE_FIRST;

		for (j = 0 ; j < IE_LINES && i < len ; i++, j++) {
			if (isel == i)
				wattron(iwin, ATTR_CURSEL);
			snprintf(tmp, sizeof(tmp), "%08lx\t%s", ie[i].discid, ie[i].title);
			mvwprintw(iwin, y++, 5, "%-*s", longl + 9, tmp);
			wrefresh(iwin);
			wattrset(iwin, 0);
		}

		switch (wgetch(iwin)) {
			case KEY_UP:
			case KEY_LEFT:
			case '=':
			case '+':
				if (isel - 1 >= 0 && isel - 1 < len)
					--isel;
				break;
			case KEY_DOWN:
			case KEY_RIGHT:
			case '-':
			case '_':
				if (isel + 1 >= 0 && isel + 1 < len)
					++isel;
				break;
			case '\n':
				xstrncpy(genre, ie[isel].genre, n);
				wclear(iwin);
				wrefresh(iwin);
				delwin(iwin);
				return (ie[isel].discid);
			case 'q':
				wclear(iwin);
				wrefresh(iwin);
				delwin(iwin);
				return (0);
		}
	}

	/* NOTREACHED */
	return (0);
}
