/*
** ocd - ojnk cd player
** Copyright (C) 1999 Odin <odin@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
*/

#include <ocd.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <signal.h>
#include <pwd.h>
#include <sys/time.h>

#define TR_LINES (max_y - 11)
#define IGNORE_PL 0x01
#define IGNORE_RC 0x02

#define S_NONE 0
#define S_PLAY 1
#define S_STOP 2
#define S_PLAYT 3

WINDOW *mwin;
int max_x, max_y;
u_char sel, startup_track = 0, startup_mode = NORMAL, startup_cmd = S_NONE;

static void help_screen(void);
static void print_version(void);
static int ocdrc_defaults(struct cdrom *cdrom);
static void die(int signal);
static void kill_cd(struct cdrom *cdrom);
static void next_track(struct cdrom *cdrom);
static void prev_track(struct cdrom *cdrom);
static void shuffle_tracks(struct plist **plist, disc_t *disc);

int main(int argc, char **argv) {
	int opt, c;
	u_int fl = 0;
	u_char temp, r_track = 0, t_sec, t_min;
	short oldvol = -1;
	u_char *d = DEFAULT_CDROM_DEV, gotobuf[96];
	struct cdrom cdrom;
#ifdef NCURSES_MOUSE_VERSION
	MEVENT mouse;
#endif

	signal(SIGSEGV, die);
	signal(SIGFPE, die);
	signal(SIGILL, die);
	signal(SIGQUIT, die);
	signal(SIGINT, die);
	signal(SIGTERM, die);
	signal(SIGHUP, SIG_IGN);

	memset(&cdrom, 0, sizeof(cdrom));

	while ((opt = getopt(argc, argv, "d:hiqv")) != EOF) {
		switch (opt) {
			case 'd':
				d = optarg;
				break;
			case 'i':
				fl |= IGNORE_PL;
				break;
			case 'h':
				help_screen();
				/* exits */
			case 'q':
				fl |= IGNORE_RC;
				break;
			case 'v':
				print_version();
				exit(0);
			default:
				help_screen();
		}
	}

	xstrncpy(cdrom.device, DEFAULT_CDROM_DEV, sizeof(cdrom.device));
	xstrncpy(cdrom.cddb.server, CDDB_HOST, sizeof(cdrom.cddb.server));
	xstrncpy(cdrom.cddb.sserver, CDDB_SHOST, sizeof(cdrom.cddb.sserver));
	xstrncpy(cdrom.cddb.dir, CDDB_DIR, sizeof(cdrom.cddb.dir));
	cdrom.cddb.port = CDDB_PORT;
	cdrom.cddb.sport = CDDB_SPORT;

	if (!(fl & IGNORE_RC))
		temp = ocdrc_defaults(&cdrom);

	if (cdrom.cddb.dir[0] != '/') {
		u_char *home = getenv("HOME");
		u_char *dtemp = xmalloc(sizeof(cdrom.cddb.dir));

		strcpy(dtemp, cdrom.cddb.dir);

		if (home == NULL) {
			struct passwd *pw = getpwuid(getuid());

			if (pw == NULL) {
				fprintf(stderr, "Unable to determine home directory.\n");
				exit(-1);
			}
			home = pw->pw_dir;
		}
		snprintf(cdrom.cddb.dir, sizeof(cdrom.cddb.dir), "%s/%s", home, dtemp);
		free(dtemp);
	}

	cdrom.fd = ocd_setup(cdrom.device, &cdrom.disc);

	if (cdrom.fd == -1) {
		fprintf(stderr, "Error: No disc in the drive.\n");
		exit(-1);
	}

/*
** See if ~/.cddb (or whatever cdrom.cddb.dir is) exists, if it's not,
** create it if possible.
*/
	if (check_dir(cdrom.cddb.dir) == -1) {
		fprintf(stderr, "Cannot create: %s\n", cdrom.cddb.dir);
		exit(-1);
	}

	ocd_cdinfo(cdrom.fd, &cdrom.disc);
	ocd_update(cdrom.fd, &cdrom.disc);
	if (temp == 1)
		ocd_setvol(cdrom.fd, &cdrom.vol);
	ocd_getvol(cdrom.fd, &cdrom.vol);	

	if (cdrom.disc.status == CDROM_AUDIO_NO_STATUS)
		ocd_stop(cdrom.fd);

	mwin = initscr();

	if (mwin == NULL) {
		fprintf(stderr, "Fatal: initscr()\n");
		exit(-1);
	}

	setup_colors();
	wbkgdset(mwin, COLOR_PAIR(13));
	wclear(mwin);
	wrefresh(mwin);
	cbreak();
	noecho();
	curs_set(FALSE);
	keypad(mwin, TRUE);
	halfdelay(10);
	getmaxyx(mwin, max_y, max_x);

	opt = cddb_lookup(&cdrom);
	if (opt == -1) {
		endwin();
		fprintf(stderr, "Weirdness\n");
		exit(-1);
	} else if (opt == 1 && !(fl & IGNORE_PL)) {
		if (cdrom.disc.status != CDROM_AUDIO_PLAY) {
			ocd_stop(cdrom.fd);
			cdrom.disc.status = CDROM_AUDIO_COMPLETED;
			cdrom.mode = PLAYLIST;
		}
	}

#ifdef NCURSES_MOUSE_VERSION
	mousemask(BUTTON1_CLICKED, NULL);
#endif

	mvwprintw(mwin, max_y - 1, 0, "Initializing...");
	wrefresh(mwin);

	if (!(fl & IGNORE_RC)) {
		switch (startup_mode) {
			case SHUFFLE:
				shuffle_tracks(&cdrom.cddb.playorder, &cdrom.disc);
				cdrom.mode = SHUFFLE;
				break;
			case PLAYLIST:
				cdrom.mode = PLAYLIST;
				break;
			case REP_T:
				cdrom.mode = REP_T;
				r_track = startup_track;
				break;
			case REP_C:
				cdrom.mode = REP_C;
		}

		switch (startup_cmd) {
			case S_PLAYT:
				ocd_play(cdrom.fd, startup_track);
				break;
			case S_PLAY:
				next_track(&cdrom);
				break;
			case S_STOP:
				ocd_stop(cdrom.fd);
				break;
		}
	}

	for (sel = 1 ;;) {
	loop:
		if (cdrom.disc.opened != 1) {
			switch (cdrom.mode) {
				case REP_T:
					/* FIXME: this is slightly broken */
					if (cdrom.disc.c_track != r_track
						|| cdrom.disc.status == CDROM_AUDIO_COMPLETED)
							ocd_play_track(cdrom.fd, r_track);
					break;
				case REP_C:
					if (cdrom.disc.status == CDROM_AUDIO_COMPLETED)
						ocd_play(cdrom.fd, 1);
					break;
				case SHUFFLE:
				case PLAYLIST:
					if (cdrom.disc.status == CDROM_AUDIO_COMPLETED)
						next_track(&cdrom);
					break;
			}
		}

		if (cdrom.disc.opened == 1 && ocd_update(cdrom.fd, &cdrom.disc) == 1) {
			cdrom.disc.track = ocd_alloctracks(cdrom.fd);
			cdrom.disc.opened = 0;
			ocd_cdinfo(cdrom.fd, &cdrom.disc);
			ocd_update(cdrom.fd, &cdrom.disc);
			ocd_stop(cdrom.fd);
			cddb_clear(&cdrom.cddb);
			cddb_lookup(&cdrom);
			sel = 1;
		}

		main_window(&cdrom);

		c = wgetch(mwin);
		switch (c) {
			/* Refresh screen */
			case '\f':
				wclear(mwin);
				break;
			/* Move down the list of tracks */
			case KEY_DOWN:
			case ' ':
				if (cdrom.disc.opened == 1)
					break;
				if (INVALID_TRACK(sel + 1, cdrom.disc.tracks))
					break;
				while (cdrom.disc.track[sel].data == 1) {
					if (!INVALID_TRACK(sel + 2, cdrom.disc.tracks))
						sel += 2;
					goto loop;
				}
				++sel;
				break;
			/* Move up the list of tracks */
			case KEY_UP:
				if (cdrom.disc.opened == 1)
					break;
				if (INVALID_TRACK(sel - 1, cdrom.disc.tracks))
					break;
				 while (cdrom.disc.track[sel - 2].data == 1) {
					if (!INVALID_TRACK(sel - 2, cdrom.disc.tracks))
						sel -= 2;
					goto loop;
				}
				--sel;
				break;
			/* Move to bottom of track list */
			case KEY_END:
				if (cdrom.disc.opened == 1)
					break;
				sel = cdrom.disc.tracks;
				break;
			/* Move to start of track list */
			case KEY_HOME:
				if (cdrom.disc.opened == 1)
					break;
				sel = 1;
				break;
			/* Page Up, up 5 tracks */
			case KEY_PPAGE:
				if (cdrom.disc.opened == 1)
					break;
				if (INVALID_TRACK(sel - 5, cdrom.disc.tracks))
					sel = 1;
				else
					sel -= TR_LINES;
				break;
			/* Page down, down 5 tracks */
			case KEY_NPAGE:
				if (cdrom.disc.opened == 1)
					break;
				if (INVALID_TRACK(sel + 5, cdrom.disc.tracks))
					sel = cdrom.disc.tracks;
				else
					sel += TR_LINES;
				break;
			/* Play selected track */
			case '\n':
				if (cdrom.disc.opened == 1)
					break;
				ocd_play(cdrom.fd, sel);
				if (cdrom.mode == REP_T)
					r_track = sel;
				break;
			/* Play next track */
			case '=':
			case '+':
			play_next:
				if (cdrom.disc.opened == 1)
					break;
				next_track(&cdrom);
				if (cdrom.mode == REP_T) {
					ocd_update(cdrom.fd, &cdrom.disc);
					r_track = cdrom.disc.c_track;
				}
				break;
			/* Play previous track */
			case '-':
			play_prev:
				if (cdrom.disc.opened == 1)
					break;
				prev_track(&cdrom);
				if (cdrom.mode == REP_T) {
					ocd_update(cdrom.fd, &cdrom.disc);
					r_track = cdrom.disc.c_track;
				}
				break;
			/* Mute, toggles */
			case '0':
			mute:
				if (oldvol == -1) {
					oldvol = cdrom.vol.channel0;
					memset(&cdrom.vol, 0, sizeof(cdrom.vol));
				} else {
					cdrom.vol.channel0 = (u_char) oldvol;
					cdrom.vol.channel1 = cdrom.vol.channel0;
					cdrom.vol.channel2 = cdrom.vol.channel0;
					cdrom.vol.channel3 = cdrom.vol.channel0;
					oldvol = -1;
				}
				ocd_setvol(cdrom.fd, &cdrom.vol);
				ocd_getvol(cdrom.fd, &cdrom.vol);
				break;
			/* Raise volume */
			case '>':
			case '.':
			raise_vol:
				oldvol = -1;
				if (cdrom.vol.channel0 > 250)
					cdrom.vol.channel0 = 255;
				else
					cdrom.vol.channel0 += 5;

				cdrom.vol.channel1 = cdrom.vol.channel0;
				cdrom.vol.channel2 = cdrom.vol.channel0;
				cdrom.vol.channel3 = cdrom.vol.channel0;
				
				ocd_setvol(cdrom.fd, &cdrom.vol);
				ocd_getvol(cdrom.fd, &cdrom.vol);
				break;
			/* Lower volume */
			case '<':
			case ',':
			lower_vol:
				oldvol = -1;
				if (cdrom.vol.channel0 < 5)
					cdrom.vol.channel0 = 0;
				else
					cdrom.vol.channel0 -= 5;

				cdrom.vol.channel1 = cdrom.vol.channel0;
				cdrom.vol.channel2 = cdrom.vol.channel0;
				cdrom.vol.channel3 = cdrom.vol.channel0;

				ocd_setvol(cdrom.fd, &cdrom.vol);
				ocd_getvol(cdrom.fd, &cdrom.vol);
				break;
			/* Close tray */
			case 'C':
			case 'c':
				ocd_close(cdrom.fd);
				break;
			/* CDDB window */
			case 'D':
			case 'd':
			cddb_win:
				if (cdrom.disc.opened == 1)
					break;
				cddb_control(&cdrom);
				break;
			/* Eject CD */
			case 'E':
			case 'e':
			eject:
				if (ocd_eject(cdrom.fd) != 0)
					break;
				if (cdrom.disc.track != NULL)
					kill_cd(&cdrom);
				cdrom.disc.opened = 1;
				break;
			/* Forward 10 seconds */
			case 'F':
			case 'f':
			case KEY_RIGHT:
			ff:
				if (cdrom.disc.opened == 1)
					break;
				if (cdrom.disc.status == CDROM_AUDIO_PLAY) {
					ocd_play_pos(cdrom.fd, cdrom.disc.c_track,
						cdrom.disc.t_pos.sec + cdrom.disc.t_pos.min * 60 + 10);
				}
				if (cdrom.mode == REP_T) {
					ocd_update(cdrom.fd, &cdrom.disc);
					r_track = cdrom.disc.c_track;
				}
				break;
			/* Play the given track */
			case 'G':
			case 'g':
			gt:
				if (cdrom.disc.opened == 1)
					break;
				user_input(mwin, 4, 11, 14, gotobuf, sizeof(gotobuf) - 1);
				temp = (u_char) atoi(gotobuf);
				if (temp < 1 || temp > cdrom.disc.tracks)
					break;
				if (cdrom.cddb.playorder != NULL) {
					destroy_playlist(cdrom.cddb.playorder);
					cdrom.cddb.playorder = NULL;
					cdrom.mode = NORMAL;
				}
				ocd_play(cdrom.fd, temp);
				break;
			/* Repeat (loop) Track */
			case 'l':
			rep:
				if (cdrom.disc.opened == 1)
					break;
				if (cdrom.disc.status == CDROM_AUDIO_PLAY) {
					r_track = cdrom.disc.c_track;
					cdrom.mode = REP_T;
				}
				break;
			/* Repeat (loop) CD */
			case 'L':
				if (cdrom.disc.opened == 1)
					break;
				cdrom.mode = REP_C;
				break;
			/* Playlist */
			case 'O':
			case 'o':
				if (cdrom.disc.opened == 1)
					break;
				boldword(mwin, max_y - 3, 3, "E", "nter a playlist", ":", 6);
				wmove(mwin, max_y - 2, 2);
				wclrtoeol(mwin);
				user_input(mwin, max_y - 2, 2, 0, gotobuf, sizeof(gotobuf));
				wrefresh(mwin);
				if (strlen(gotobuf) < 1)
					break;
				if (parse_playlist(gotobuf, &cdrom.cddb) != -1) {
					ocd_stop(cdrom.fd);
					cdrom.mode = PLAYLIST;
					cdrom.disc.status = CDROM_AUDIO_COMPLETED;
				}
				break;
			/* Pause or resume the cd */
			case 'P':
			case 'p':
			pause:
				if (cdrom.disc.opened == 1)
					break;
				if (cdrom.disc.status == CDROM_AUDIO_PAUSED) {
					ocd_resume(cdrom.fd);
					break;
				} else if (cdrom.disc.status == CDROM_AUDIO_PLAY)
					ocd_pause(cdrom.fd);
				break;
			/* Quit */
			case 'Q':
			case 'q':
			quit:
				wclear(mwin);
				wrefresh(mwin);
				endwin();
				exit(0);
			/* Rewind 10 seconds */
			case 'R':
			case 'r':
			case KEY_LEFT:
			rw:
				if (cdrom.disc.opened == 1)
					break;
				if (cdrom.disc.status == CDROM_AUDIO_PLAY) {
					ocd_play_pos(cdrom.fd, cdrom.disc.c_track,
						cdrom.disc.t_pos.sec + cdrom.disc.t_pos.min * 60 - 10);
				}
				if (cdrom.mode == REP_T) {
					ocd_update(cdrom.fd, &cdrom.disc);
					r_track = cdrom.disc.c_track;
				}
				break;
			/* Stop CD */
			case 'S':
			case 's':
			stop:
				if (cdrom.disc.opened == 1)
					break;
				if (cdrom.mode == PLAYLIST || cdrom.mode == SHUFFLE) {
					if (cdrom.cddb.playorder != NULL) {
						destroy_playlist(cdrom.cddb.playorder);
						cdrom.cddb.playorder = NULL;
					}
				}
				cdrom.mode = NORMAL;
				ocd_stop(cdrom.fd);
				break;
			/* Go to a specific time of the track or the disc */
			case 'T':
			case 't':
				if (cdrom.disc.opened == 1)
					break;
				if (c == 't' && cdrom.disc.c_track == 0)
					break;

				user_input(mwin, (c == 'T' ? 6 : 5), 11, 6,
					gotobuf, MIN(sizeof(gotobuf) - 1, 5));

				d = strchr(gotobuf, ':');
				if (d == NULL)
					break;

				*d++ = '\0';
				t_min = (u_char) atoi(gotobuf);
				t_sec = (u_char) atoi(d);

				if (c == 'T')
					ocd_play_pos(cdrom.fd, 1, t_min * 60 + t_sec);
				else {
					ocd_play_pos(cdrom.fd, cdrom.disc.c_track,
						t_min * 60 + t_sec);
				}
				break;
			/* Shuffle */
			case 'x':
			case 'X':
				if (cdrom.disc.opened == 1)
					break;
				shuffle_tracks(&cdrom.cddb.playorder, &cdrom.disc);
				cdrom.mode = SHUFFLE;
				ocd_stop(cdrom.fd);
				break;
#ifdef NCURSES_MOUSE_VERSION
			case KEY_MOUSE:
				getmouse(&mouse);
				if (mouse.x >= (COL_1 - 2) && mouse.x <= (COL_2 - 1)) {
					switch (mouse.y) {
						case CY_NEXT:
							goto play_next;
						case CY_PREV:
							goto play_prev;
						case CY_STOP:
							goto stop;
						case CY_PAUSE:
							goto pause;
						case CY_REW:
							goto rw;
						case CY_FF:
							goto ff;
						break;
					}
				} else if (mouse.x >= (COL_2 - 2) && mouse.x <= (COL_3 - 1)) {
					switch (mouse.y) {
						case CY_EJECT:
							goto eject;
						case CY_CLOSE:
							break;
						case CY_GOTO:
							goto gt;
						case CY_VOLR:
							goto raise_vol;
						case CY_VOLL:
							goto lower_vol;
						case CY_MUTE:
							goto mute;
						break;
					}
				} else if (mouse.x >= (COL_3 - 2) && mouse.x <= (max_x - 2)) {
					switch (mouse.y) {
						case CY_CDDB:
							goto cddb_win;
						case CY_REP:
							goto rep;
						case CY_MODE:
							break;
						case CY_QUIT:
							goto quit;
						break;
					}
				}
				break;
#endif
		}
	}

	endwin();
	exit(0);
}

static void shuffle_tracks(struct plist **plist, disc_t *disc) {
	u_int i;
	u_char ntracks, temp, j, k, tracks[disc->tracks];
	struct timeval tv;

	ntracks = disc->tracks;

	destroy_playlist(*plist);
	*plist = NULL;

	for (i = 0 ; i < ntracks ; i++) {
		if (disc->track[i].data == 1)
			ntracks--;
		else
			tracks[i] = i + 1;
	}

	gettimeofday(&tv, NULL);
	srandom(tv.tv_sec ^ (tv.tv_usec << 11));

	/* Overkill */
	for (i = 0 ; i < 1000 ; i++) {
		for (j = 0 ; j < ntracks ; j++) {
			k = random() / (RAND_MAX / ntracks + 1);
			temp = tracks[j];
			tracks[j] = tracks[k];
			tracks[k] = temp;
		}
	}
	
	for (i = 0 ; i < ntracks ; i++)
		append_node(tracks[i], plist);

	while ((*plist)->prev != NULL)
		*plist = (*plist)->prev;
}

static void kill_cd(struct cdrom *cdrom) {
	if (cdrom->disc.track != NULL)
		xfree(cdrom->disc.track);

	destroy_playlist(cdrom->cddb.playorder);

	if (cdrom->cddb.track != NULL)
		xfree(cdrom->cddb.track);

	if (cdrom->cddb.playorder != NULL)
		xfree(cdrom->cddb.playorder);

	cdrom->cddb.discid = 0;
	cdrom->cddb.tracks = 0;
	cdrom->cddb.llen = 0;
	cdrom->disc.min = 0;
	cdrom->disc.sec = 0;
	cdrom->disc.tracks = 0;
	memset(cdrom->cddb.title, 0, sizeof(cdrom->cddb.title));
	memset(cdrom->cddb.ext, 0, sizeof(cdrom->cddb.ext));
}

static void next_track(struct cdrom *cdrom) {
	switch (cdrom->mode) {
		case REP_C:
			if (cdrom->disc.c_track == cdrom->disc.tracks)
				ocd_play(cdrom->fd, 1);
			else
				ocd_play(cdrom->fd, cdrom->disc.c_track + 1);
			break;
		case REP_T:
		case NORMAL:
			if (cdrom->disc.c_track == cdrom->disc.tracks)
				ocd_play(cdrom->fd, cdrom->disc.c_track);
			else if (!INVALID_TRACK(cdrom->disc.c_track+ 1, cdrom->disc.tracks))
				ocd_play(cdrom->fd, cdrom->disc.c_track + 1);
			break;
		case SHUFFLE:
		case PLAYLIST:
			if (cdrom->cddb.playorder->next != NULL) {
				ocd_play_track(cdrom->fd, cdrom->cddb.playorder->track);
				cdrom->cddb.playorder = cdrom->cddb.playorder->next;
			}
			break;
	}
}

static void prev_track(struct cdrom *cdrom) {
	switch (cdrom->mode) {
		case REP_C:
			if (cdrom->disc.c_track == 1)
					ocd_play(cdrom->fd, cdrom->disc.tracks);
			else
				ocd_play(cdrom->fd, cdrom->disc.c_track - 1);
			break;
		case REP_T:
		case NORMAL:
			if (cdrom->disc.c_track == 1)
				ocd_play(cdrom->fd, 1);
			else if (!INVALID_TRACK(cdrom->disc.c_track - 1, cdrom->disc.tracks))
				ocd_play(cdrom->fd, cdrom->disc.c_track - 1);
			break;
		case SHUFFLE:
		case PLAYLIST:
			if (cdrom->cddb.playorder->prev != NULL) {
				ocd_play_track(cdrom->fd, cdrom->cddb.playorder->track);
				cdrom->cddb.playorder = cdrom->cddb.playorder->prev;
			}
			break;
	}
}

static void die(int signal) {
	endwin();
	fprintf(stderr, "\nPOW! Signal %d\n", signal);
	exit(-1);
}

static int ocdrc_defaults(struct cdrom *cdrom) {
	u_char ge, l = 0, buf[1024], temp[1024], *val, *item, *r;
	FILE *fp;
	int volchange = 0;

	r = getenv("HOME");
	if (r == NULL) {
		struct passwd *pw = getpwuid(getuid());
		if (pw == NULL) {
			fprintf(stderr, "You don't exist.\n");
			exit(-1);
		}
		snprintf(temp, sizeof(temp), "%s/%s", pw->pw_dir, OCDRC);
	} else
		snprintf(temp, sizeof(temp), "%s/%s", r, OCDRC);

	fp = fopen(temp, "r");
	if (fp == NULL)
		return (0);

	while (ocd_getline(fp, buf, sizeof(buf) - 1) != EOF) {
		r = buf;
		while (*r == ' ' || *r == '\t')
			r++;

		if (*r == '#' || *r == '\0')
			continue;

		for (ge = 0, item = temp ; *r != '\0' ; ) {
			switch (*r) {
				case '\n':
					break;
				case '\t':
					*r = ' ';
					/* FALLTHROUGH */
				case ' ':
					if (ge == 0 || l == ' ' || l == '=') {
						r++;
						break;
					}
					/* FALLTHROUGH */
				case '=':
					ge = 1;
					/* FALLTHROUGH */
				default:
					*item++ = l = *r++;
			}
		}

		*item = '\0';
		if (*--item == ' ')
			*item = '\0';

		val = strchr(temp, '=');
		if (val == NULL)
			continue;
		*val++ = '\0';

		switch (temp[0]) {
			case 'd':
				xstrncpy(cdrom->device, val, sizeof(cdrom->device));
				break;
			case 'v':
				r = strchr(val, '%');
				cdrom->vol.channel0 = (u_char) strtoul(val, NULL, 10);
				if (r != NULL) {
					cdrom->vol.channel0 =
						(u_char) ocd_rint(cdrom->vol.channel0 / 100.0 * 255.0);
				}
				cdrom->vol.channel1 = cdrom->vol.channel0;
				cdrom->vol.channel2 = cdrom->vol.channel0;
				cdrom->vol.channel3 = cdrom->vol.channel0;
				volchange = 1;
				break;
			case 'c':
				if (!strcasecmp(temp, "cddb_dir"))
					xstrncpy(cdrom->cddb.dir, val, sizeof(cdrom->cddb.dir));
				else if (!strcasecmp(temp, "cddb_server")) {
					xstrncpy(cdrom->cddb.server, val,
						sizeof(cdrom->cddb.server));
				} else if (!strcasecmp(temp, "cddb_sserver")) {
					xstrncpy(cdrom->cddb.sserver, val,
						sizeof(cdrom->cddb.sserver));
				} else if (!strcasecmp(temp, "cddb_port")) {
					cdrom->cddb.port = strtoul(val, NULL, 10);
				} else if (!strcasecmp(temp, "cddb_sport"))
					cdrom->cddb.sport = strtoul(val, NULL, 10);
				break;
			case 's':
				if (!strcasecmp(temp, "startup_command")) {
					if (!strncasecmp(val, "stop", 4))
						startup_cmd = S_STOP;
					else if (!strncasecmp(val, "play_track", 10)) {
						r = strchr(val, ' ');
						if (r == NULL)
							break;
						startup_track = strtoul(++r, NULL, 10);
						startup_cmd = S_PLAYT;
					} else if (!strncasecmp(val, "play", 4))
						startup_cmd = S_PLAY;
				} else if (!strncasecmp(temp, "startup_mode", 12)) {
					if (!strcasecmp(val, "repeat_cd"))
						startup_mode = REP_C;
					else if (!strncasecmp(val, "repeat_track", 12)) {
						r = strchr(val, ' ');
						if (r == NULL)
							break;
						startup_track = strtoul(++r, NULL, 10);
						startup_mode = REP_T;
					} else if (!strcasecmp(val, "shuffle"))
						startup_mode = SHUFFLE;
				} else if (!strncasecmp(val, "playlist", 8)) {
					r = strchr(val, ' ');
					if (r == NULL)
						break;
					if (parse_playlist(++r, &cdrom->cddb) != 1)
						startup_mode = PLAYLIST;
				}
				break;
		}
	}

	fclose(fp);
	return (volchange);
}

static void print_version(void) {
	fprintf(stderr, "ocd v%s.%s\n", VERSION, REVISION);
}

static void help_screen(void) {
	print_version();
	fprintf(stderr,
		"\t-d <device>\tcdrom device\n"
		"\t-i\t\tignore the playorder cddb keyword\n"
		"\t-h\t\tthis help screen\n"
		"\t-q\t\tignore .ocdrc\n"
		"\t-v\t\tdisplay the version\n"
	);

	exit(0);
}
