/*
** ocd - ojnk cd player
** Copyright (C) 1999 Odin <odin@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
*/

#include <ocd.h>
#include <limits.h>
#include <sys/stat.h>
#include <pwd.h>
#include <netdb.h>
#include <ctype.h>
#include <sys/socket.h>
#include <arpa/inet.h>

/* Same as strncpy, only this nul-terminates the string */
char *xstrncpy(char *dest, const char *src, size_t n) {
	u_char *ret = dest;

	if (n == 0)
		return (dest);

	while (--n > 0 && (*dest++ = *src++) != '\0')
		;
	*dest = '\0';

	return (ret);
}

/* Only append n chars to dest if room, nul-terminate. */
char *xstrncat(char *dest, const char *src, size_t n) {
	u_char *ret = dest;
	
	if (n == 0)
		return (dest);

	for (; *dest != '\0' ; dest++, n--)
		;
	while (--n > 0 && (*dest++ = *src++) != '\0')
		;
	*dest = '\0';

	return (ret);
}

void *xcalloc(size_t nmemb, size_t size) {
	void *p = calloc(nmemb, size);

	if (p == NULL) {
		endwin();
		fprintf(stderr, "Fatal: Out of memory.\n");
		exit(-1);
	}

	return (p);
}

void *xmalloc(size_t size) {
	void *p = malloc(size);

	if (p == NULL) {
		endwin();
		fprintf(stderr, "Fatal: Out of memory.\n");
		exit(-1);
	}

	return (p);
}

#ifndef __GLIBC__
int dprintf(int fd, const char *fmt, ...) {
	u_char buf[BUFSIZ];
	va_list ap;

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);

	return (write(fd, buf, strlen(buf)));
}

int vdprintf(int fd, const char *fmt, va_list ap) {
	u_char buf[BUFSIZ];

	vsnprintf(buf, sizeof(buf), fmt, ap);

	return (write(fd, buf, strlen(buf)));
}
#endif

int check_dir(const u_char *dir) {
	if (access(dir, W_OK | X_OK) == -1) {
		if (errno == ENOENT)
			return (mkdir(dir, 0700));
		return (-1);
	}

	return (0);
}

int ocd_rint(double orig) {
    /* adequate for our purpose.. */
    if (orig - (int) orig >= .50)
        return (((int) orig) + 1);
    else
        return ((int) orig);
}

int get_addr(const u_char *hostname, u_long *addr) {
	struct in_addr in;

	if (inet_aton(hostname, &in)) {
		*addr = in.s_addr;
		return (0);
	} else {
		struct hostent *host = gethostbyname(hostname);
		if (host) {
			*addr = *((u_long *) host->h_addr);
			return (0);
		}
	}

	return (-1);
}

int host_connect(const u_char *host, int port) {
	int sock;
	u_long addr;
	struct sockaddr_in sin;

	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock == -1)
		return (-1);

	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_port = htons(port);

	if (get_addr(host, &addr) == -1) {
		close(sock);
		return (-1);
	}

	sin.sin_addr.s_addr = addr;

	if (connect(sock, (struct sockaddr *) &sin, sizeof(sin)) == -1) {
		close(sock);
		return (-1);
	}

	return (sock);
}

int ocd_getline(FILE *fp, u_char *buf, size_t len) {
	int c;
	size_t i = 0;

	for (;;) {
		switch ((c = getc(fp))) {
			case EOF:
				if (i == 0)
					return EOF;
			case '\n':
				if (i >= len)
					i = 0;
				buf[i] = '\0';
				return (0);
			default:
				if (i == 0 && isspace(c))
					break;
				if (i < len)
					buf[i++] = c;
		}
	}

	return (0);
}

ssize_t sock_write(int sock, void *buf, size_t len) {
	ssize_t n;

	while (len > 0) {
		n = write(sock, buf, len);
		if (n == -1)
			return (-1);
		len -= n;
		buf += n;
	}

	return (len);
}

ssize_t sock_readline(int fd, u_char *buf, size_t len) {
	u_int i;
	u_char c;
	ssize_t ret;

	for (i = 1 ; i < len ; i++) {
		ojnk:
			ret = read(fd, &c, 1);
			if (ret == 1) {
				if (buf != NULL)
					*buf++ = c;
				if (c == '\n')
					break;
			} else if (ret == 0) {
				if (i == 1)
					return (0);
				else
					break;
			} else {
				if (errno == EINTR)
					goto ojnk;
				return (-1);
			}
	}

	if (buf != NULL)
		*buf = '\0';

	return (i);
}

int astrcmp(const u_char *const orig, const u_char *const *const w, size_t n) {
	u_int i, j;
	size_t size = n * sizeof(u_char *);
	u_char **test = malloc(size);

	memcpy(test, w, size);

	for (i = 0 ; ; i++) {
		if (i > 0) {
			if (orig[i - 1] == '\0')
				break;
		}
		for (j = 0 ; j < n ; j++) {
			if (test[j] != NULL) {
				if (toupper(orig[i]) != toupper(test[j][i]))
					test[j] = NULL;
				else if (orig[i] == '\0' && test[j][i] == '\0') {
					free(test);
					return (j);
				}
				
			}
		}
	}

	free(test);
	return (-1);
}

void destroy_playlist(struct plist *p) {
	if (p != NULL) {
		while (p->prev != NULL)
			p = p->prev;

		while (p->next != NULL) {
			p = p->next;
			free(p->prev);
		}

		free(p);
	}
}

/* simple list append... mallocs space for the first node if necessary */
struct plist *append_node(u_char track, struct plist **p) {
	if (*p == NULL) {
		*p = xmalloc(sizeof(struct plist));
		(*p)->next = NULL;
		(*p)->prev = NULL;
		(*p)->track = track;
		return (*p);
	}

	while ((*p)->next != NULL)
		*p = (*p)->next;

	(*p)->next = xmalloc(sizeof(struct plist));
	(*p)->next->prev = *p;
	(*p)->next->next = NULL;
	(*p)->next->track = track;

	return ((*p)->next);
}
